  /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */

package Java2OWL;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.generic.BasicType;
import org.apache.bcel.util.BCELifier;


/** This is a collection of useful static methods.
 *
 * @author ohlbach
 */
public class Utilities {
    
    /** for loading classes from given files. */
    private static FileClassLoader classLoader = new FileClassLoader();
    
    /** maps class names of primitive classes to wrapper classes */
    private static final Map<String,Class> toWrapper = new HashMap<String,Class>();
    static {
            toWrapper.put("byte",   Byte.class);
            toWrapper.put("short",  Short.class);
            toWrapper.put("char",   Character.class);
            toWrapper.put("int",    Integer.class);
            toWrapper.put("long",   Long.class);
            toWrapper.put("float",  Float.class);
            toWrapper.put("double", Double.class);
            toWrapper.put("boolean", Boolean.class);
        }
    
    /** returns a return class of the method which can me mapped to OWL.
     * <p>
     * If the return class is a composite class (array or collection) then 
     * a suitable component type is returned. 
     * Only primitive types, Wrapper-classes, or other non-composite classes are suitable.
     * Component types of arrays and of collections must also be a non-composite class.
     * 
     * @param method any Java-method
     * @return the return class or its component type respectively, or null if it cannot be mapped to OWL.
     */
    static Class getReturnClass(Method method) {
        return getComponentClass(method.getReturnType(),method.getGenericReturnType());}
    
    
     /** returns a component type which can be mapped to OWL.
     * @param cl typically a parameter type of a method
     * @param ty typically a generic parameter type of a method.
     * <p>
     * If the class is a composite class (array or collection) then 
     * a suitable component type is returned. 
     * Only primitive types, Wrapper-classes, or other non-composite classes are suitable.
     * Component types of arrays and of collections must also be a non-composite class.
     * 
     * @return the class or its component type respectively, or null if it cannot be mapped to OWL.
     */
    static Class getComponentClass(Class cl, Type ty) {
        if(cl == void.class) return null;
        if(cl.isPrimitive()) return cl;
        if(cl.isArray()) {
            Class acl = cl.getComponentType();
            if(acl.isArray()) return null;  // arrays of arrays can not be used.
            if(acl.isPrimitive()) {return acl;}
            Type aty;
            try{aty = ((GenericArrayType)ty).getGenericComponentType();}
            catch(ClassCastException e) {return acl;}
            if(acl == aty) return acl;
            return null;}
        if(cl == ty) {return cl;}
        try{cl.asSubclass(Collection.class);} catch(ClassCastException e) {return null;}
        // now we check collection types
        Type[] types = ((ParameterizedType)ty).getActualTypeArguments();
        if(types.length > 1) return null;
        try{Class rcl = (Class)(types[0]);
            if(rcl.isArray()) return null;
            return rcl;}
        catch(ClassCastException e) {return null;} }
    
     /** checks whether the methods returns a map 
     * 
     * @param method any method
     * @return [type1,type2] if the method returns a map, otherwise null
     * <p> Example: HashMap<String,Integer> getMap() 
     * yields  [String,Integer]
     */
    public static Type[] returnsMap(Method method) {
        if(Map.class.isAssignableFrom(method.getReturnType())) {
            ParameterizedType ty = (ParameterizedType)(method.getGenericReturnType());
            return ty.getActualTypeArguments();}
        else return null;}
    
    
    /** returns true if the return type is such that it can take multiple values.
     * 
     * @param method any Java-method.
     * @return true if the return type is a composite type whose component type can be mapped to OWL.
     */
    static boolean isRelational(Method method) {
        Class cl = method.getReturnType(); 
        Type ty = method.getGenericReturnType();
        if(cl.isPrimitive()) return false;
       
        if(cl.isArray()) {
            Class acl = cl.getComponentType();
            if(acl.isArray()) return false;
            if(acl.isPrimitive()) {return true;}
            Type aty;
            try{aty = ((GenericArrayType)ty).getGenericComponentType();}
            catch(ClassCastException e) {return true;}
            if(acl == aty) return true;
            return false;}
        if(cl == ty) {return false;}
        try{cl.asSubclass(Collection.class);} catch(ClassCastException e) {return false;}
        
        Type[] types = ((ParameterizedType)ty).getActualTypeArguments();
        if(types.length > 1) return false;
        try{Class rcl = (Class)(types[0]);
            if(rcl.isArray()) return false;
            return true;}
        catch(ClassCastException e) {return false;}}
    
    /** returns true if the Java data type corresponds to an OWL data type.
     * <p>
     * This is essentially a string, a boolean or a number type (primitive or wrapper).
     * @param jcl a Java-datatype (class).
     * @return true if it can be mapped to an OWL-datatype.
     */    
    static boolean isOWLDatatype(Class jcl) {
        if(jcl.isPrimitive()) return true;
        if(jcl == String.class) return true;  
        if(jcl == Boolean.class) return true;
        if(Number.class.isAssignableFrom(jcl)) return true;
        return false;}
    
    /** checks whether the types denote one-dimensional arrays of ordinary objects */
     static Class isArrayType(Class jcl, Type type) {
        if(!jcl.isArray()) return null;
        Type componentType = null;
        try{GenericArrayType ar = (GenericArrayType)type;
            componentType = ar.getGenericComponentType();}
        catch(Exception ex) {
            Class componentClass = jcl.getComponentType();
            return componentClass.isArray() ? null :componentClass ;}
        return (isClass(componentType) != null) ? jcl.getComponentType() : null;}
    
    /** returns a string generated by concatenating the given string s n-times. */
    public static String concatenate(String s, int n) {
        String r = "";
        for(int i = 0; i < n; ++i) r += s;
        return r;} 
            
    /** adjusts lines consisting of pairs separated by the given separator.
     * 
     * <p>
     * The resulting string should be formatted as a proper table. 
     * @param text      a string consisting of lines like, for example,  abc -> cde.
     * @param separator a string separating pairs in the line.
     * @return          the string with the first element of each pair filled with blanks.
     */
    public static String adjustPairs(String text, String separator) {
        String[] lines = text.split("\n");
        int l = 0;
        for(String line : lines) {
            String[] pair = line.split(separator);
            if(pair.length > 1)
                l = (l < pair[0].length()) ? pair[0].length() : l;}
        String result = "";
        for(String line : lines) {
            String[] pair = line.split(separator);
            if(pair.length == 1) result += concatenate(" ",l+separator.length()) + line + "\n";
            else {
                result += pair[0] + concatenate(" ", l - pair[0].length());
                for(int i = 1; i < pair.length; ++i) result += separator + pair[i] + "\n";}}
        return result.substring(0,result.length()-1);}
    
    /** This method can be used to parse argument strings of main methods.
     * 
     * @param parameters this is typically the argument list of a main method.
     * @param postfixes a list of postfixes to be matched by the parameters.
     * @param keys      a list of keys for key,value pairs.
     * @return An array with three elements:<br/>
     * 1. {@code HashMap<String,ArrayList<String>> postFixParameters} the recognised parameters matching the postfixes.<br/>
     * 2. {@code HashMap<String,String> keyValues}  the recognised values of the key-value pairs.<br/>
     * 3. {@code ArrayList<String> rest}            all the unrecognised parameters.
     * 
     * <p>
     * Example for a postfix parameter ".class" to be matched with all java-class files
     * For the key-value pairs, it is assumed that they occur in a sequence "key value" in the parameters. 
     */
    public static Object[] parseParameters(String[] parameters, Set<String> postfixes, Set<String> keys) {
        HashMap<String,ArrayList<String>> postFixParameters = new HashMap<String,ArrayList<String>>();
        HashMap<String,String> keyValues = new HashMap<String,String>();
        ArrayList<String> rest = new ArrayList<String>(); 
                
        Map<String,Boolean> keyFlags = new HashMap<String,Boolean>();
        for(String key : keys) keyFlags.put(key, Boolean.FALSE);
        
        for(String postfix : postfixes) {
            ArrayList<String> list = new ArrayList<String>();
            postFixParameters.put(postfix,list);}
        
        for(String parameter : parameters) {
            boolean found = false;
            for(String postfix : postfixes) {
                if(parameter.endsWith(postfix)) {
                    postFixParameters.get(postfix).add(parameter);
                    found = true;
                    break;}}
            if(found) continue;
            
            for(Entry<String,Boolean>  keyFlag : keyFlags.entrySet()) {
                if(keyFlag.getValue() == Boolean.TRUE) {
                    keyValues.put(keyFlag.getKey(), parameter);
                    keyFlags.put(keyFlag.getKey(), Boolean.FALSE);
                    found = true;
                    break;}}
            if(found) continue;
            
            if(keys.contains(parameter)) {
                keyFlags.put(parameter,Boolean.TRUE);
                continue;}
            rest.add(parameter);}
        return new Object[]{postFixParameters,keyValues,rest};
    }
    
    /** returns a string describing all environment variables. */
    public static String getEnvirentmentVariables () {
        String s = "";
        for(Entry entry : System.getenv().entrySet()) 
            s += entry.getKey() + " = " + entry.getValue() + "\n";
        return adjustPairs(s,"=");}
    
    /** returns a string describing all system properties. */
    public static String getSystemProperties() {
        String s = "";
        for(Entry entry : System.getProperties().entrySet()) 
            s += entry.getKey() + " = " + entry.getValue() + "\n";
        return adjustPairs(s,"=");}
    
    /** loads a compiled Java-class from the given file.
     * 
     * <p>
     * If the class has already been loaded, the loaded class is returned.
     * @param file the compiled Java-file.
     * @return the class or null if a read-error occurred or it is an interface.
     * @throws ClassNotFoundException.
     */
    public static Class classFromFile(File file) throws  ClassNotFoundException {
        String classname = file.getAbsolutePath();
        if(classname.endsWith(".class")) classname = classname.substring(0,classname.length()-6);
        Class jcl = null;
        String[] parts = classname.split(System.getProperties().getProperty("file.separator"));
        classname = parts[parts.length-1];
        
        // we check whether the class is already there.
        for(int i = parts.length-2; i >= 0; --i) {
            try{jcl = Class.forName(classname);} catch(ClassNotFoundException ex) {}
            if(jcl != null) return jcl;
            classname = parts[i]+"."+classname;}
        
        // now we load it from the file.
        try{jcl = classLoader.loadClass(file.getAbsolutePath());}
        catch(Error er) {throw new ClassNotFoundException(er.getMessage(),er);}
        return jcl;}
    
       
    /** loads all compiled Java-classes in the given directory.
     * 
     * @param directory the directory with the compiled Java-files.
     * @return [classes,error-string] Interfaces are not included.
     */
    public static Object[]classesFromDirectory(File directory) { 
        ArrayList<Class> classes = new ArrayList<Class>();
        String errors = "";
        String blanks = "     ";
        Class jcl = null;
        for(File file : directory.listFiles(
            new FilenameFilter(){
                    @Override
                    public boolean accept(File f,String s) {return s.endsWith(".class");}})) {
            try {jcl = classFromFile(file);}
            catch (Exception ex) {
                errors += blanks + file + " " + ex.getMessage() + "\n";continue;}
            if(jcl != null) classes.add(jcl);}
        if(!errors.isEmpty()) errors = "Errors when loading class files from directory " + directory +":\n" + errors;
        return new Object[]{classes,errors};}
    
    /** applies the function to the object or the components of an array or a collection. 
     * 
     * <p>
     * If the object is an array or a collection, then the function is applied to all of its elements.<br/>
     * The application is stopped as soon as the {@code apply}-function returns false. <br/>
     * The {@code apply}-function is tried in the following sequence:<br/>
     * 1. if object is an array,<br/>
     * 2. if object is a collection,<br/>
     * 3. otherwise it is just applied to the object itself. 
     * 
     * @param object an array, a collection or just an object. 
     * @param function a function to be applied at the elements of the array or collection or the object itself
     * @return true if all applications succeeded.
     */
    public static boolean map(Object object, Function function) {
        if(object == null) return true;
        if(object.getClass().isArray()) {
            int length = Array.getLength(object);
            for(int i = 0; i < length; ++i) {
                if(!function.apply(Array.get(object,i))) return false;}
            return true;}
        
        if(object instanceof Collection) {
            for(Object element : (Collection)object) {
                if(!function.apply(element)) return false;}
            return true;}
        
        return function.apply(object);}
       
    /** finds the stackframe just above the last one whose class starts with the given prefix.
     * <p>
     * Classes starting with java, sun., org.,  are ignored.
     * 
     * @param prefix The prefix of a fully qualified class name.
     * @return the stackframe above the first occurrence of this class or null if none could be found.
     */
    public static StackTraceElement findStackFrame(String prefix) {
        boolean lookingForPrefix = true;
        for (StackTraceElement frame : new Throwable().getStackTrace()) {
            String cname = frame.getClassName(); 
            boolean prefixFound = cname.startsWith(prefix);
            if (lookingForPrefix) {
                     // Skip all frames until we have found the first logger frame.
                    if (prefixFound) {lookingForPrefix = false; }}
            else {
                if (!prefixFound) {
                         // skip reflection call
                    if (!(cname.startsWith("java") || cname.startsWith("sun.") || 
                            cname.startsWith("org."))) {
                        // We've found the relevant frame.
                        return frame;}}}}
        return null;}
    
    /** returns the result of printStackTrace() as a string. */
    public static String stackTraceString(Throwable throwable) {
        StringWriter buffer = new StringWriter(); 
        throwable.printStackTrace(new PrintWriter(buffer));
        return buffer.toString();}
           
    /** applies a method (with empty argument list) to an object.
     * 
     * <p>
     * This works for cases where {@code object.<methodname>()} does not work
     * because it is unknown how to cast the object.
     *
     * @param object a Java-object
     * @param methodname the name of a method without arguments
     * @return the result of invoking the method to the object.
     * @throws Exception if the invocation failed. 
     */
    public static Object invoke0(Object object, String methodname) throws Exception {
        return object.getClass().getMethod(methodname).invoke(object);}
            
    /** returns the name of the object. 
     * <p> First it tries a getName() method, then the toString() method.
     */
    public static String getName(Object object) {
        try {return (String)invoke0(object,"getName");}
        catch(Exception ex) {return object.toString();}}
    
    /** sets the Boolean value of a static field.
     * 
     * @param jcl a Java class
     * @param name the name of a static field
     * @param value the boolean value to be set.
     * @throws Exception if the operation failed.
     */
    public static void setStaticField(Class jcl, String name, boolean value) throws Exception {
        jcl.getField(name).setBoolean(null, value);}
    
    /** sets the value of a static field.
     * @param jcl a Java class
     * @param name the name of a static field
     * @param value the boolean value to be set.
     * @throws Exception if the operation failed.
     */
    public static void setStaticField(Class jcl, String name, Object value) throws Exception {
        jcl.getField(name).set(null, value);}
    
    /** sets the value into an object field.
     * @param object a Java object.
     * @param name the name of an instance field
     * @param value the value to be set.
     * @throws Exception if the operation failed.
     */
    public static void setObjectField(Object object, String name, Object value) throws Exception {
        object.getClass().getField(name).set(object, value);} 
  
    /** sets the flag in the object field to 1.
     * 
     * @param object an object with an int field named 'name'
     * @param name the name of this field.
     * @param flag a position in a word, e.g. 1,2,4,8,16,...
     * @throws Exception if the operation failed.
     */
    public static void setObjectFlag(Object object, String name, int flag) throws Exception {
        Field field = object.getClass().getField(name);
        field.setInt(object, field.getInt(object) | flag);} 
    
    /** returns true if the flag in the object field is 1.
     * 
     * @param object an object with an int field named 'name'
     * @param name the name of this field.
     * @param flag a position in a word, e.g. 1,2,4,8,16,...
     * @throws Exception if the operation failed.
     */
    public static boolean getObjectFlag(Object object, String name, int flag) throws Exception {
        return (object.getClass().getField(name).getInt(object) & flag) != 0;} 
   
    /** sets the flag in the object field to 0.
     * 
     * @param object an object with an int field named 'name'
     * @param name the name of this field.
     * @param flag a position in a word, e.g. 1,2,4,8,16,...
     * @throws Exception if the operation failed.
     */
    public static void resetObjectFlag(Object object, String name, int flag) throws Exception {
        Field field = object.getClass().getField(name);
        field.setInt(object, field.getInt(object) & ~flag);} 
    
    
    /** sets the flag in the object field to 0, but returns the previous value.
     * 
     * @param object an object with an int field named 'name'
     * @param name the name of this field.
     * @param flag a position in a word, e.g. 1,2,4,8,16,...
     * @return true if the flag was 1.
     * @throws Exception if the operation failed.
     */
    public static boolean getAndResetObjectFlag(Object object, String name, int flag) throws Exception {
        Field field = object.getClass().getField(name);
        int value = field.getInt(object);
        field.setInt(object, value & ~flag);
        return (value & flag) != 0;}
        
   
     /** sets the flag in the object field to 1, but returns the previous value.
     * 
     * @param object an object with an int field named 'name'
     * @param name the name of this field.
     * @param flag a position in a word, e.g. 1,2,4,8,16,...
     * @return true if the flag was 1.
     * @throws Exception if the operation failed.
     */
    public static boolean getAndSetObjectFlag(Object object, String name, int flag) throws Exception {
        Field field = object.getClass().getField(name);
        int value = field.getInt(object);
        field.setInt(object, value | flag);
        return (value & flag) != 0;}
        
   
    
    /** turns the input parameters into an ArrayList. 
     * <p> Notice that this method is not generic because objects of mixed types are to be put into the list.  
     */
    public static List<Object> asList(Object... objects) {return Arrays.asList(objects);}
    
    /** prints the current stack trace (for testing purposes). */
    public static void printStack() {
        Throwable t = new Throwable();
        t.printStackTrace();}
    
    /** returns for a primitive type the corresponding wrapper type name. */ 
    public static String getWrapperType(BasicType type) {
       String name = type.toString();
       if(name.equals("int")) name = "integer";
       return "java.lang." + name.substring(0,1).toUpperCase() + name.substring(1);}
    
    /** returns for a primitive Java-type the BCEL BasicType. */
    public static BasicType getBasicType(Class type) {
        String name = type.getName();
        for(byte b = 4; b < 12; ++b) {
            BasicType bt = BasicType.getType(b);
            if(bt.toString().equals(name)) return bt;}
        return null;}
    
    /** BCELifies the class and sends it to the file or System.out.
     * 
     * @param jcl a Java class
     * @param filename a filename or null
     * @throws ClassNotFoundException
     * @throws FileNotFoundException 
     */
    public static void bcelify(Class jcl, String filename) throws ClassNotFoundException, FileNotFoundException {
        JavaClass jclb = Repository.lookupClass(jcl);
        PrintStream out = (filename == null) ? System.out : new PrintStream(filename); 
        BCELifier bclf = new BCELifier(jclb,out);
        bclf.start();}
    
    /** transfers the values of fromObject's non-static fields to the corresponding fields of the toObject. 
     * 
     * @param fromObject the sending Java object.
     * @param toObject the receiving Java object. 
     * @param toClass the to-object's class.
     * <p> All member-fields, public, private, protected, inherited are transferred. 
     * Only fields protected by a special security manager cannot be transferred. 
     * For a particular field F in the to-object, the method goes top-down in the 
     * inheritance hierarchy to find a field with the same name and type as F.
     * This field is transferred. No deep copy is made!
     */
    public static void transferFields(Object fromObject, Object toObject, Class toClass) {
        Class superclass = toClass.getSuperclass();
        if(superclass != null) transferFields(fromObject,toObject,superclass);
        for(Field toField  : toClass.getDeclaredFields()) {
            if((toField.getModifiers() & Modifier.STATIC) != Modifier.STATIC) {
                Field fromField = findField(fromObject.getClass(),toField.getName(),toField.getType());
                if(fromField != null) transferField(fromObject,fromField,toObject,toField);}}}
    
    /** transfers the contents of fromField to toField.
     * 
     * @param fromObject          the source object
     * @param fromField     the source field
     * @param toObject            the target object
     * @param toField       the target field
     * <p> If the field is protected by a security manager, it is not transferred. 
     */
     private static void transferField(Object fromObject, Field fromField, Object toObject, Field toField) {
         try {
             toField.setAccessible(true);
             fromField.setAccessible(true);
             if(toField.isAccessible() && fromField.isAccessible()) toField.set(toObject, fromField.get(fromObject));}
         catch (Exception ex) { // should never happen
               ex.printStackTrace();
               System.exit(1);}
         finally{toField.setAccessible(false);
                 fromField.setAccessible(false);}}
     
    /** This method transfers the contents of the common non-static fields from fromObjects[0],...,fromObjects[i-1] to toObject.
      * 
      * @param fromObjects     an array of objects.
      * @param i               an index into fromObjects
      * @param toObject        an object.
      * @param toClass         the toObject's class. 
      * 
      * <p>For each field F in toObject, it searches fromObjects[0],...,fromObjects[i-1]
      * to find a field with the same name and type. The contents of the first one it finds is transferred 
      * to F.
      * <p> If the field is protected by a security manager, it is not transferred. 
      * <p>Fields in toObjects which have already a non-null value are kept as they are.
      */
    public static void transferCommonFields(Object[] fromObjects, int i, Object toObject, Class toClass) {
        Class superclass = toClass.getSuperclass();
        if(superclass != null) transferCommonFields(fromObjects,i,toObject,superclass);
        for(Field toField  : toClass.getDeclaredFields()) {
            if((toField.getModifiers() & Modifier.STATIC) != Modifier.STATIC) {
                try{Object value = toField.get(toObject);
                    if(!isNull(value)) continue;} catch(Exception ex) {}
                for(int j = 0; j < i; ++j) {
                    Object fromObject = fromObjects[j];
                    Field fromField = findField(fromObject.getClass(),toField.getName(),toField.getType());
                    if(fromField != null){
                        transferField(fromObject,fromField,toObject,toField);
                        break;}}}}}
    
    /** searches for a field in object which has the given name and type.
     * 
     * @param jcl    the object's class or superclass.
     * @param name   a field name.
     * @param type   a field type.
     * @return the object's field with the same name and type like the given field, or null if there is none of this type.
     * 
     * <p>
     * The method must search recursively through the object's superclasses because getDeclaredFields()
     * returns only the fields declared in the given class.
     */
    
    private static  Field  findField(Class jcl, String name, Type type) {
         Class superclass = jcl.getSuperclass();
         Field field = null;
         if(superclass != null) field = findField(superclass,name,type);
         if(field != null) return field;
         try {field = jcl.getDeclaredField(name);} catch (Exception ex) {return null;}
         return field.getType().equals(type) ? field : null;}
       
    /** returns true if the first class is a wrapper class for the second one. */
    static boolean isBoxable(Class wrapper, Class primitive) {
        if(!primitive.isPrimitive()) return false;
        String wName = wrapper.getSimpleName();
        String pName = primitive.getSimpleName();
        if(wName.equals("Boolean") && pName.equals("boolean")) return true;
        if(!(Number.class.isAssignableFrom(wrapper))) return false;
        // wrapper is now really a wrapper class for numbers
        if(wName.toLowerCase().equals(pName)) return true;
        return wName.equals("Integer") && pName.equals("int");
    }
    
    /** replaces all whitespace characters by c */
    static String replaceWhitespace(String s, char c) {
        String[] parts = s.split("\\s");
        String result = parts[0];
        for(int i = 1; i < parts.length; ++i) result += c + parts[i];
        return result;}
    
    /** returns true if the object has the 0 value. */
    static Boolean isNull(Object object) {
        if(object == null) return true;
        if(object.hashCode() == 0) return true;
        if(object instanceof Boolean && (Boolean)object == false) return true;
        if(object.getClass().isArray() && Array.getLength(object) == 0) return true;
        if(object instanceof Collection && ((Collection)object).size() == 0) return true;
        return false;}
    
    /** searches a constructor with a given argument class
     * 
     * @param jcl       the class containting the constructor.
     * @param argument  the class of the first argument of the constructor. 
     * @return          the constructor or null if none could be found.
     */
    static public Constructor findConstructor(Class jcl, Class argument) { 
        for(Constructor constructor : jcl.getDeclaredConstructors()) {
            Class[] parameters = constructor.getParameterTypes();
            if(parameters.length == 1 && argument == parameters[0]) {return constructor;}}
        return null;}
    
    /** transfers a set into a collection of the given targetClass
     * 
     * @param items         a set
     * @param targetClass   a collection class
     * @return              a collection containing the items, or null if an error occurred.
     */
    static public Object transferSet(Set items, Class targetClass) {
        if(items.getClass() == targetClass) return items;
        if(targetClass.isArray()) {
            Object arr = Array.newInstance(targetClass.getComponentType(), items.size());
            int i = -1;
            for(Object ob : items) {Array.set(arr,++i,ob);}
            return arr;}
        
        Object collection = null;
        try {
            collection = targetClass.newInstance();
            Method adder = targetClass.getMethod("add", Object.class);
            for(Object item : items) adder.invoke(collection,item);}
            catch(Exception ex) {return null;}
        return collection;}
    
    /** checks whether the given type is an array type.
     * 
     * @param t any type
     * @return the corresponding array class or null if it is not an array.
     */
    static public Class isArray(Type t) {
        try{return ((Class)t).isArray() ? (Class)t : null;}
        catch(Exception ex) {return null;}}
    
    
     /** checks whether the given type is an ordinary class, no array and no parameterized type.
     * 
     * @param t any type
     * @return the corresponding class or null if it is not a normal class.
     */
    static public Class isClass(Type t) {
        try{return ((Class)t).isArray() ? null : (Class)t;}
        catch(Exception ex) {return null;}}
    
      /** checks whether the given type is a parameterized type.
     * 
     * @param t any type
     * @return the corresponding parameterized type or null if it is not a parameterized.
     */
    static public ParameterizedType isParameterizedType(Type t) {
        try{return (ParameterizedType)t;}
        catch(Exception ex) {return null;}}
    
    
     /** checks whether the type is an admissible collection class.
      * 
      * @param t
      * @return the corresponding class or null if it is not admissible.
      * <p>Examples: <br/>
      * HashSet<String> is admissible <br/>
      * Set<String> is not admissible because Set cannot be instantiated.
      * HashSet<String[]> is not admissible because the components must be ordinary classes<br/>
      */
     static public Class isCollection(Type t) {
        try{
            ParameterizedType tp = (ParameterizedType)t; 
            Class jcl = (Class)(tp.getRawType());
            if(Collection.class.isAssignableFrom(jcl)) { 
                Type[] arguments = tp.getActualTypeArguments();
                if(arguments.length == 1 && !(((Class)arguments[0]).isArray())) return jcl;}}
        catch(Exception ex) {}
        return null;}
     
     /** returns for a generic array type the corresponding Java class */
     static public Class getArrayClass(GenericArrayType type) {
            Type componentType = type.getGenericComponentType();
            try {
                Class cc = Class.forName(componentType.toString().substring(6));
                Object ar = Array.newInstance(cc, 0);
                return ar.getClass();}
            catch(Exception ex) {return null;}}
    
    
    
    /** The main method prints the environment variables and the system properties. */
    public static void main(String[] args) {
        System.out.println("Environment Variables");
        System.out.println(getEnvirentmentVariables());
        System.out.println("\nSystem Properties");
        System.out.println(getSystemProperties());
    }
    
}
